# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
This top level module contains the main qBraid QIR functionality.

.. currentmodule:: qbraid_qir

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   dumps

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   QbraidQirError
   QirConversionError

"""
from ._version import __version__
from .cirq import cirq_to_qir
from .exceptions import QbraidQirError, QirConversionError
from .serialization import dumps



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing unit tests for Cirq to QIR conversion functions.

"""
# isort: skip_file

import cirq
import pytest

from qbraid_qir.cirq.convert import cirq_to_qir
from tests.fixtures.basic_gates import (
    double_op_tests,
    measurement_tests,
    rotation_tests,
    single_op_tests,
    triple_op_tests,
)

from .qir_utils import assert_equal_qir
from .test_utils import (
    check_attributes,
    double_op_call_string,
    generic_op_call_string,
    get_entry_point_body,
    initialize_call_string,
    measure_call_string,
    result_record_output_string,
    return_string,
    rotation_call_string,
    single_op_call_string,
)


def test_cirq_to_qir_type_error():
    """Test raising exception for bad input type."""
    with pytest.raises(TypeError):
        cirq_to_qir(None)


def test_cirq_to_qir_conversion_error():
    """Test raising exception for conversion error."""
    circuit = cirq.Circuit()
    with pytest.raises(ValueError):
        cirq_to_qir(circuit)


@pytest.mark.parametrize("circuit_name", single_op_tests)
def test_single_qubit_gates(circuit_name, request):
    qir_op, circuit = request.getfixturevalue(circuit_name)
    qir_module = cirq_to_qir(circuit, record_output=False)
    qir_str = str(qir_module).splitlines()
    func = get_entry_point_body(qir_str)
    assert func[0] == initialize_call_string()
    assert func[1] == single_op_call_string(qir_op, 0)
    assert func[2] == return_string()
    assert len(func) == 3


@pytest.mark.parametrize("circuit_name", rotation_tests)
def test_rotation_gates(circuit_name, request):
    qir_op, circuit = request.getfixturevalue(circuit_name)
    generated_qir = str(cirq_to_qir(circuit)).splitlines()
    check_attributes(generated_qir, 1, 1)
    func = get_entry_point_body(generated_qir)
    assert func[0] == initialize_call_string()
    assert func[1] == rotation_call_string(qir_op, 0.5, 0)
    assert func[3] == return_string()
    assert len(func) == 4


@pytest.mark.parametrize("circuit_name", double_op_tests)
def test_double_qubit_gates(circuit_name, request):
    qir_op, circuit = request.getfixturevalue(circuit_name)
    generated_qir = str(cirq_to_qir(circuit)).splitlines()
    check_attributes(generated_qir, 2, 2)
    func = get_entry_point_body(generated_qir)
    assert func[0] == initialize_call_string()
    assert func[1] == double_op_call_string(qir_op, 0, 1)
    assert func[4] == return_string()
    assert len(func) == 5


@pytest.mark.parametrize("circuit_name", triple_op_tests)
def test_triple_qubit_gates(circuit_name, request):
    qir_op, circuit = request.getfixturevalue(circuit_name)
    generated_qir = str(cirq_to_qir(circuit)).splitlines()
    check_attributes(generated_qir, 3, 3)
    func = get_entry_point_body(generated_qir)
    assert func[0] == initialize_call_string()
    assert func[1] == generic_op_call_string(qir_op, [0, 1, 2])
    assert func[5] == return_string()
    assert len(func) == 6


@pytest.mark.parametrize("circuit_name", measurement_tests)
def test_measurement(circuit_name, request):
    qir_op, circuit = request.getfixturevalue(circuit_name)
    module = cirq_to_qir(circuit)
    generated_qir = str(module).splitlines()
    check_attributes(generated_qir, 1, 1)
    func = get_entry_point_body(generated_qir)

    assert func[0] == initialize_call_string()
    assert func[1] == measure_call_string(qir_op, 0, 0)
    assert func[2] == result_record_output_string(0)
    assert func[3] == return_string()
    assert len(func) == 4


def test_verify_qir_bell_fixture(pyqir_bell):
    """Test that pyqir fixture generates code equal to test_qir_bell.ll file."""
    assert_equal_qir(pyqir_bell.ir(), "test_qir_bell")


def test_entry_point_name(cirq_bell):
    """Test that entry point name is consistent with module ID."""
    name = "quantum_123"
    module = cirq_to_qir(cirq_bell, name=name)
    assert module.source_filename == name


def test_convert_bell_compare_file(cirq_bell):
    """Test converting Cirq bell circuit to QIR."""
    test_name = "test_qir_bell"
    module = cirq_to_qir(
        cirq_bell, name=test_name, initialize_runtime=False, record_output=False
    )
    assert_equal_qir(str(module), test_name)


@pytest.mark.skip(reason="Test case incomplete")
def test_qft():
    """Test converting Cirq QFT circuit to QIR."""
    for n in range(2, 5):  # Test for different numbers of qubits
        circuit = cirq.Circuit()
        qubits = [cirq.NamedQubit(f"q{i}") for i in range(n)]
        circuit.append(cirq.qft(*qubits))
        # TODO Add assertions or checks here



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for processing Cirq circuits before conversion to QIR.

"""
import itertools
from typing import List

import cirq

from qbraid_qir.cirq.opsets import map_cirq_op_to_pyqir_callable
from qbraid_qir.exceptions import QirConversionError


def _decompose_gate_op(op: cirq.GateOperation) -> List[cirq.OP_TREE]:
    try:
        # Try converting to PyQIR. If successful, keep the operation.
        _ = map_cirq_op_to_pyqir_callable(op)
        return [op]
    except QirConversionError:
        pass
    new_ops = cirq.decompose_once(op, flatten=True, default=[op])
    if len(new_ops) == 1 and new_ops[0] == op:
        return [op]
    return list(itertools.chain.from_iterable(map(_decompose_gate_op, new_ops)))


def _decompose_unsupported_gates(circuit: cirq.Circuit) -> cirq.Circuit:
    """
    Decompose gates in a circuit that are not in the supported set.

    Args:
        circuit (cirq.Circuit): The quantum circuit to process.

    Returns:
        cirq.Circuit: A new circuit with unsupported gates decomposed.
    """
    new_circuit = cirq.Circuit()
    for moment in circuit:
        new_ops = []
        for op in moment:
            if isinstance(op, cirq.GateOperation):
                decomposed_ops = _decompose_gate_op(op)
                new_ops.extend(decomposed_ops)
            else:
                new_ops.append(op)

        new_circuit.append(new_ops)
    return new_circuit


def preprocess_circuit(circuit: cirq.Circuit) -> cirq.Circuit:
    """
    Preprocesses a Cirq circuit to ensure that it is compatible with the QIR conversion.

    Args:
        circuit (cirq.Circuit): The Cirq circuit to preprocess.

    Returns:
        cirq.Circuit: The preprocessed Cirq circuit.

    """
    qubit_map = {
        qubit: cirq.LineQubit(i) for i, qubit in enumerate(circuit.all_qubits())
    }
    line_qubit_circuit = circuit.transform_qubits(lambda q: qubit_map[q])
    return _decompose_unsupported_gates(line_qubit_circuit)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing unit tests for CirqModule and Module elements.

"""
import hashlib

import cirq

from qbraid_qir.cirq.elements import generate_module_id


def test_generate_module_id_format(cirq_bell):
    """Test generating module ID for a Cirq circuit fits expected format."""
    module_id = generate_module_id(cirq_bell)
    assert module_id.startswith("circuit-")
    assert len(module_id) == 15


def test_generate_module_id_hex(cirq_bell):
    """Test if generated module ID is consistent with manual SHA-256 computation."""
    module_id = generate_module_id(cirq_bell)
    serialized_circuit = cirq.to_json(cirq_bell)
    hash_object = hashlib.sha256(serialized_circuit.encode())
    hash_hex = hash_object.hexdigest()
    alphanumeric_hash = "".join(filter(str.isalnum, hash_hex))
    truncated_hash = alphanumeric_hash[:7]

    expected_id = f"circuit-{truncated_hash}"
    assert module_id == expected_id



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for containing QIR code utils functions used for unit tests.

"""
import os


def assert_equal_qir(given_qir: str, filename: str) -> None:
    """Function that compares generated qir to the qir in a file.

    Args:
        given_qir (str): Given qir string that should be compared with the file.
        filename (str): Name of the file that should be compared with the given qir.
    """
    current_dir = os.path.dirname(__file__)

    resources_file = os.path.join(current_dir, "resources", f"{filename}.ll")

    with open(resources_file, encoding="utf-8") as f:
        file_data = f.read().strip()

    processed_given_qir = given_qir.strip()

    assert file_data == processed_given_qir



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module mapping supported Cirq gates/operations to pyqir functions.

"""
from typing import Callable, Tuple

import cirq
import pyqir._native

from qbraid_qir.exceptions import QirConversionError

# NOTE: Upper/lower case matters here, and were set to
# match the Cirq gate/op string representation.

PYQIR_OP_MAP = {
    # Single-Qubit Clifford Gates
    "H": pyqir._native.h,
    "X": pyqir._native.x,
    "Y": pyqir._native.y,
    "Z": pyqir._native.z,
    # Single-Qubit Rotation Gates
    "Rx": pyqir._native.rx,
    "Ry": pyqir._native.ry,
    "Rz": pyqir._native.rz,
    # Single-Qubit Non-Clifford Gates
    "S": pyqir._native.s,
    "T": pyqir._native.t,
    "S**-1": pyqir._native.s_adj,
    "T**-1": pyqir._native.t_adj,
    # Two-Qubit Gates
    "SWAP": pyqir._native.swap,
    "CNOT": pyqir._native.cx,
    "CZ": pyqir._native.cz,
    # Three-Qubit Gates
    "TOFFOLI": pyqir._native.ccx,
    # Classical Gates/Operations
    "MEASURE": pyqir._native.mz,
    "reset": pyqir._native.reset,
}


def map_cirq_op_to_pyqir_callable(op: cirq.Operation) -> Tuple[Callable, str]:
    """
    Maps a Cirq operation to its corresponding PyQIR callable function.

    Args:
        op (cirq.Operation): The Cirq operation to map.

    Returns:
        Tuple[Callable, str]: Tuple containing the corresponding PyQIR callable function,
                               and a string representing the gate/operation type.

    Raises:
        QirConversionError: If the operation or gate is not supported.
    """
    if isinstance(op, cirq.ops.GateOperation):
        gate = op.gate

        if isinstance(gate, cirq.ops.MeasurementGate):
            op_name = "MEASURE"
        elif isinstance(gate, (cirq.ops.Rx, cirq.ops.Ry, cirq.ops.Rz)):
            op_name = gate.__class__.__name__
        elif isinstance(gate, cirq.ops.Pauli):
            op_name = gate.__class__.__name__[-1]  # X, Y, Z
        elif isinstance(
            gate, (cirq.ops.XPowGate, cirq.ops.YPowGate, cirq.ops.ZPowGate)
        ):
            if gate.exponent == 1 or (
                isinstance(gate, cirq.ZPowGate)
                and gate.exponent in [0.25, -0.25, 0.5, -0.5]
            ):
                op_name = str(gate)  # X, Y, Z, S, T, S**-1, T**-1
            else:
                op_name = f"R{gate.__class__.__name__[0].lower()}"  # Rotations
        else:
            op_name = str(gate)

    else:
        op_name = str(op)

    try:
        return PYQIR_OP_MAP[op_name], op_name
    except KeyError as err:
        raise QirConversionError(f"Cirq gate {op} not supported.") from err



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module containing unit tests Cirq to QIR utility/helper functions

"""
# isort: skip_file

from typing import List

from pyqir import (
    Context,
    Function,
    Module,
    is_entry_point,
    required_num_qubits,
    required_num_results,
)


def _qubit_string(qubit: int) -> str:
    if qubit == 0:
        return "%Qubit* null"
    return f"%Qubit* inttoptr (i64 {qubit} to %Qubit*)"


def _result_string(res: int) -> str:
    if res == 0:
        return "%Result* null"
    return f"%Result* inttoptr (i64 {res} to %Result*)"


def initialize_call_string() -> str:
    return "call void @__quantum__rt__initialize(i8* null)"


def single_op_call_string(name: str, qb: int) -> str:
    return f"call void @__quantum__qis__{name}__body({_qubit_string(qb)})"


def double_op_call_string(name: str, qb1: int, qb2: int) -> str:
    return f"call void @__quantum__qis__{name}__body({_qubit_string(qb1)}, {_qubit_string(qb2)})"


def rotation_call_string(name: str, theta: float, qb: int) -> str:
    return f"call void @__quantum__qis__{name}__body(double {theta:#e}, {_qubit_string(qb)})"


def measure_call_string(name: str, res: str, qb: int) -> str:
    return f"call void @__quantum__qis__{name}__body({_qubit_string(qb)}, {_result_string(res)})"


def array_record_output_string(num_elements: int) -> str:
    return (
        f"call void @__quantum__rt__array_record_output(i64 {num_elements}, i8* null)"
    )


def result_record_output_string(res: str) -> str:
    return f"call void @__quantum__rt__result_record_output({_result_string(res)}, i8* null)"


def generic_op_call_string(name: str, qbs: List[int]) -> str:
    args = ", ".join(_qubit_string(qb) for qb in qbs)
    return f"call void @__quantum__qis__{name}__body({args})"


def return_string() -> str:
    return "ret void"


def get_entry_point(mod: Module) -> Function:
    func = next(filter(is_entry_point, mod.functions))
    assert func is not None, "No main function found"
    return func


def get_entry_point_body(qir: List[str]) -> List[str]:
    joined = "\n".join(qir)
    mod = Module.from_ir(Context(), joined)
    func = next(filter(is_entry_point, mod.functions))
    assert func is not None, "No main function found"
    lines = str(func).splitlines()[2:-1]
    return list(map(lambda line: line.strip(), lines))


def check_attributes_on_entrypoint(
    func: Function, expected_qubits: int = 0, expected_results: int = 0
) -> None:
    actual_qubits = -1
    actual_results = -1

    actual_qubits = required_num_qubits(func)
    actual_results = required_num_results(func)
    assert (
        expected_qubits == actual_qubits
    ), f"Incorrect qubit count: {expected_qubits} expected, {actual_qubits} actual"

    assert (
        expected_results == actual_results
    ), f"Incorrect result count: {expected_results} expected, {actual_results} actual"


def check_attributes(
    qir: List[str], expected_qubits: int = 0, expected_results: int = 0
) -> None:
    x = "\n".join(qir)
    mod = Module.from_ir(Context(), x)
    func = next(filter(is_entry_point, mod.functions))

    check_attributes_on_entrypoint(func, expected_qubits, expected_results)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for exporting QIR to bitcode and LLVM IR files.

"""
import logging
import os
import pathlib
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    import pyqir


def dumps(module: "pyqir.Module", output_dir: Optional[str] = None) -> None:
    """
    Saves the Quantum Intermediate Representation (QIR) of a given module
    to bitcode (.bc) and LLVM IR (.ll) files.

    Args:
        module: The module containing the QIR.
        output_dir (Optional[str]): The directory where the files will be saved.
                                    If not provided, defaults to the current
                                    working directory.

    Raises:
        Exception: If there's an error in writing the files.
    """
    # Determine the filename prefix from the module's source filename
    filename_prefix = os.path.splitext(os.path.basename(module.source_filename))[0]

    # Set the output directory to the current working directory if not provided
    file_dir = output_dir if output_dir else str(pathlib.Path.cwd())

    # Create the directory if it doesn't exist
    os.makedirs(file_dir, exist_ok=True)

    # Construct file paths
    bc_file = os.path.join(file_dir, f"{filename_prefix}.bc")
    ll_file = os.path.join(file_dir, f"{filename_prefix}.ll")

    try:
        # Save bitcode file (does not require an encoding)
        with open(bc_file, "wb") as f:  # pylint: disable=unspecified-encoding
            f.write(module.bitcode)
        logging.info("Saved to %s", bc_file)

        # Save LLVM IR file (default python encoding)
        with open(ll_file, "w", encoding="utf-8") as f:
            f.write(str(module))
        logging.info("Saved to %s", ll_file)

    except Exception as e:
        logging.error("An error occurred: %s", e)
        raise



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Cirq circuit fixtures for unit tests.

"""
import cirq
import pytest


@pytest.fixture
def cirq_bell() -> cirq.Circuit:
    """Returns a Cirq bell circuit with measurement over two qubits."""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1), cirq.measure(q0, q1))
    return circuit



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Test functions that decompose unsupported Cirq gates before conversion to QIR.

"""

import cirq
import numpy as np
from qbraid.interface import circuits_allclose

from qbraid_qir.cirq.passes import _decompose_unsupported_gates


def test_only_supported_gates():
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(qubits[0]), cirq.CNOT(qubits[0], qubits[1]))
    decomposed_circuit = _decompose_unsupported_gates(circuit)
    assert decomposed_circuit == circuit
    assert circuits_allclose(decomposed_circuit, circuit)


def test_contains_unsupported_gates():
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.ops.ISwapPowGate(exponent=np.pi).on(*qubits),
    )
    decomposed_circuit = _decompose_unsupported_gates(circuit)
    assert decomposed_circuit != circuit
    assert circuits_allclose(decomposed_circuit, circuit)


def test_empty_circuit():
    circuit = cirq.Circuit()
    decomposed_circuit = _decompose_unsupported_gates(circuit)
    assert decomposed_circuit == circuit
    assert circuits_allclose(decomposed_circuit, circuit)


def test_custom_gate():
    class CustomGate(cirq.Gate):  # pylint: disable=abstract-method
        def _num_qubits_(self):
            return 1

        def _decompose_(self, qubits):
            yield cirq.X(qubits[0])

    custom_gate = CustomGate()
    qubit = cirq.LineQubit(0)
    circuit = cirq.Circuit(custom_gate.on(qubit))
    decomposed_circuit = _decompose_unsupported_gates(circuit)
    assert decomposed_circuit != circuit
    assert (
        any(
            isinstance(op.gate, CustomGate)
            for moment in decomposed_circuit
            for op in moment
        )
        is False
    )
    assert circuits_allclose(decomposed_circuit, circuit)


def test_multiple_decomposes():
    class CustomGate(cirq.Gate):  # pylint: disable=abstract-method
        def _num_qubits_(self):
            return 1

        def _decompose_(self, qubits):
            class InnerCustomGate(cirq.Gate):  # pylint: disable=abstract-method
                def _num_qubits_(self):
                    return 1

                def _decompose_(self, qubits):
                    yield cirq.X(qubits[0])

            inner_custom_gate = InnerCustomGate()
            yield inner_custom_gate(qubits[0])

    custom_gate = CustomGate()
    qubit = cirq.LineQubit(0)
    circuit = cirq.Circuit(custom_gate.on(qubit))
    decomposed_circuit = _decompose_unsupported_gates(circuit)
    assert decomposed_circuit != circuit
    assert (
        any(
            isinstance(op.gate, CustomGate)
            for moment in decomposed_circuit
            for op in moment
        )
        is False
    )
    assert circuits_allclose(decomposed_circuit, circuit)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Test functions that preprocess Cirq circuits before conversion to QIR.

"""
import cirq
import numpy as np
import pytest

from qbraid_qir.cirq.passes import preprocess_circuit

# pylint: disable=redefined-outer-name


@pytest.fixture
def gridqubit_circuit():
    qubits = [cirq.GridQubit(x, 0) for x in range(4)]
    circuit = cirq.Circuit(cirq.H(q) for q in qubits)
    yield circuit


@pytest.fixture
def namedqubit_circuit():
    qubits = [cirq.NamedQubit(f"q{i}") for i in range(4)]
    circuit = cirq.Circuit(cirq.H(q) for q in qubits)
    yield circuit


def test_convert_gridqubits_to_linequbits(gridqubit_circuit):
    linequbit_circuit = preprocess_circuit(gridqubit_circuit)
    for qubit in linequbit_circuit.all_qubits():
        assert isinstance(qubit, cirq.LineQubit), "Qubit is not a LineQubit"
    assert np.allclose(
        linequbit_circuit.unitary(), gridqubit_circuit.unitary()
    ), "Circuits are not equal"


def test_convert_namedqubits_to_linequbits(namedqubit_circuit):
    linequbit_circuit = preprocess_circuit(namedqubit_circuit)
    for qubit in linequbit_circuit.all_qubits():
        assert isinstance(qubit, cirq.LineQubit), "Qubit is not a LineQubit"
    assert np.allclose(
        linequbit_circuit.unitary(), namedqubit_circuit.unitary()
    ), "Circuits are not equal"


def test_empty_circuit_conversion():
    circuit = cirq.Circuit()
    converted_circuit = preprocess_circuit(circuit)
    assert (
        len(converted_circuit.all_qubits()) == 0
    ), "Converted empty circuit should have no qubits"



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Unit tests for the export module

"""
import os
import pathlib
import shutil

import pytest

from qbraid_qir.serialization import dumps

# pylint: disable=redefined-outer-name,too-few-public-methods


class MockModule:
    """Mock class for a Module object"""

    def __init__(self, source_filename, bitcode, ir_representation):
        self.source_filename = source_filename
        self.bitcode = bitcode
        self.ir_representation = ir_representation

    def __str__(self):
        return self.ir_representation


@pytest.fixture
def mock_module():
    """Fixture for creating a mock module"""
    return MockModule("mock_module", b"bitcode data", "IR representation")


@pytest.fixture
def tmp_path():
    """Provides a temporary directory for the test and cleans it up after tests are complete."""
    tmp_directory = pathlib.Path(__file__).parent / "tmp"
    tmp_directory.mkdir(exist_ok=True)
    yield tmp_directory
    shutil.rmtree(tmp_directory)


def test_save_qir_default_directory(tmp_path, mock_module):
    """Test saving QIR in the default directory"""
    os.chdir(tmp_path)
    dumps(mock_module)
    assert os.path.exists("mock_module.bc")
    assert os.path.exists("mock_module.ll")


def test_save_qir_custom_directory(tmp_path, mock_module):
    """Test saving QIR in a custom directory"""
    custom_dir = tmp_path / "custom"
    dumps(mock_module, output_dir=str(custom_dir))
    assert os.path.exists(custom_dir / "mock_module.bc")
    assert os.path.exists(custom_dir / "mock_module.ll")


def test_save_qir_with_exception(mock_module):
    """Test exception handling"""
    with pytest.raises(Exception):
        dumps(mock_module, output_dir="/non/existent/path")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining CirqVisitor.

"""
import logging
from abc import ABCMeta, abstractmethod
from typing import List

import cirq
import pyqir
import pyqir._native
import pyqir.rt
from pyqir import BasicBlock, Builder, Constant, IntType, PointerType

from qbraid_qir.cirq.elements import CirqModule
from qbraid_qir.cirq.opsets import map_cirq_op_to_pyqir_callable

_log = logging.getLogger(name=__name__)


class CircuitElementVisitor(metaclass=ABCMeta):
    @abstractmethod
    def visit_register(self, qids):
        pass

    @abstractmethod
    def visit_operation(self, operation):
        pass


class BasicQisVisitor(CircuitElementVisitor):
    """A visitor for QIS (Quantum Instruction Set) basic elements.

    This class is designed to traverse and interact with elements in a quantum circuit.

    Args:
        initialize_runtime (bool): If True, quantum runtime will be initialized. Defaults to True.
        record_output (bool): If True, output of the circuit will be recorded. Defaults to True.
    """

    def __init__(self, initialize_runtime: bool = True, record_output: bool = True):
        self._module = None
        self._builder = None
        self._entry_point = None
        self._qubit_labels = {}
        self._measured_qubits = {}
        self._initialize_runtime = initialize_runtime
        self._record_output = record_output

    def visit_cirq_module(self, module: CirqModule) -> None:
        _log.debug("Visiting Cirq module '%s' (%d)", module.name, module.num_qubits)
        self._module = module.module
        context = self._module.context
        entry = pyqir.entry_point(
            self._module, module.name, module.num_qubits, module.num_clbits
        )

        self._entry_point = entry.name
        self._builder = Builder(context)
        self._builder.insert_at_end(BasicBlock(context, "entry", entry))

        if self._initialize_runtime is True:
            i8p = PointerType(IntType(context, 8))
            nullptr = Constant.null(i8p)
            pyqir.rt.initialize(self._builder, nullptr)

    @property
    def entry_point(self) -> str:
        return self._entry_point

    def finalize(self) -> None:
        self._builder.ret(None)

    def record_output(self, module: CirqModule) -> None:
        if self._record_output is False:
            return

        i8p = PointerType(IntType(self._module.context, 8))

        for i in range(module.num_qubits):
            result_ref = pyqir.result(self._module.context, i)
            pyqir.rt.result_record_output(self._builder, result_ref, Constant.null(i8p))

    def visit_register(self, qids: List[cirq.Qid]) -> None:
        _log.debug("Visiting qids '%s'", str(qids))

        if not isinstance(qids, list):
            raise TypeError("Parameter is not a list.")

        if not all(isinstance(x, cirq.Qid) for x in qids):
            raise TypeError("All elements in the list must be of type cirq.Qid.")

        self._qubit_labels.update(
            {bit: n + len(self._qubit_labels) for n, bit in enumerate(qids)}
        )
        _log.debug("Added labels for qubits %s", str(qids))

    def visit_operation(self, operation: cirq.Operation) -> None:
        qlabels = [self._qubit_labels.get(bit) for bit in operation.qubits]
        qubits = [pyqir.qubit(self._module.context, n) for n in qlabels]
        results = [pyqir.result(self._module.context, n) for n in qlabels]

        pyqir_func, op_str = map_cirq_op_to_pyqir_callable(operation)

        if op_str == "MEASURE":
            _log.debug("Visiting measurement operation '%s'", str(operation))
            for qubit, result in zip(qubits, results):
                self._measured_qubits[pyqir.qubit_id(qubit)] = True
                pyqir_func(self._builder, qubit, result)
        elif op_str in ["Rx", "Ry", "Rz"]:
            pyqir_func(self._builder, operation.gate._rads, *qubits)
        else:
            pyqir_func(self._builder, *qubits)

    def ir(self) -> str:
        return str(self._module)

    def bitcode(self) -> bytes:
        return self._module.bitcode()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing PyQIR circuit fixtures for unit tests.

"""
import pytest
from pyqir import BasicQisBuilder, SimpleModule


@pytest.fixture
def pyqir_bell() -> SimpleModule:
    """Returns a QIR bell circuit with measurement over two qubits."""
    bell = SimpleModule("test_qir_bell", num_qubits=2, num_results=2)
    qis = BasicQisBuilder(bell.builder)

    qis.h(bell.qubits[0])
    qis.cx(bell.qubits[0], bell.qubits[1])
    qis.mz(bell.qubits[0], bell.results[0])
    qis.mz(bell.qubits[1], bell.results[1])

    return bell



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Cirq basic gate fixtures for use in tests.

"""
import cirq
import pytest

# All of the following dictionaries map from the names of methods on Cirq Circuit objects
# to the name of the equivalent pyqir BasicQisBuilder method

_zero_qubit_operations = {"barrier": "barrier"}

_one_qubit_gates = {
    "H": "h",
    "reset": "reset",
    "T": "t",
    "X": "x",
    "Y": "y",
    "Z": "z",
}

_rotations = {"Rx": "rx", "Ry": "ry", "Rz": "rz"}

_two_qubit_gates = {"CX": "cnot", "CZ": "cz", "SWAP": "swap"}

_three_qubit_gates = {"TOFFOLI": "ccx"}

_measurements = {"measure": "mz"}


def _fixture_name(s: str) -> str:
    return f"Fixture_{s}"


def _map_gate_name(gate_name: str) -> str:
    if gate_name in _one_qubit_gates:
        return _one_qubit_gates[gate_name]
    if gate_name in _measurements:
        return _measurements[gate_name]
    if gate_name in _rotations:
        return _rotations[gate_name]
    if gate_name in _two_qubit_gates:
        return _two_qubit_gates[gate_name]
    if gate_name in _three_qubit_gates:
        return _three_qubit_gates[gate_name]

    raise ValueError(f"Unknown Cirq gate {gate_name}")


def _generate_one_qubit_fixture(gate_name: str):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        q = cirq.NamedQubit("q0")
        circuit.append(getattr(cirq, gate_name)(q))
        return _map_gate_name(gate_name), circuit

    return test_fixture


# Generate simple single-qubit gate fixtures
for gate in _one_qubit_gates:
    name = _fixture_name(gate)
    locals()[name] = _generate_one_qubit_fixture(gate)


def _generate_rotation_fixture(gate_name: str):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        q = cirq.NamedQubit("q0")
        circuit.append(getattr(cirq, gate_name)(rads=0.5)(q))
        return _map_gate_name(gate_name), circuit

    return test_fixture


# Generate rotation gate fixtures
for gate in _rotations:
    name = _fixture_name(gate)
    locals()[name] = _generate_rotation_fixture(gate)


def _generate_two_qubit_fixture(gate_name: str):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        qs = [cirq.LineQubit(0), cirq.LineQubit(1)]
        circuit.append(getattr(cirq, gate_name)(qs[0], qs[1]))
        return _map_gate_name(gate_name), circuit

    return test_fixture


# Create a new function to generate a fixture for n-qubit gates
def _generate_n_qubit_fixture(gate_name: str, n: int):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        qubits = [cirq.NamedQubit(f"q{i}") for i in range(n)]
        circuit.append(getattr(cirq, gate_name)(*qubits))


# Generate double-qubit gate fixtures
for gate in _two_qubit_gates:
    name = _fixture_name(gate)
    locals()[name] = _generate_two_qubit_fixture(gate)


def _generate_three_qubit_fixture(gate_name: str):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        qs = [cirq.LineQubit(0), cirq.LineQubit(1), cirq.LineQubit(2)]
        circuit.append(getattr(cirq, gate_name)(qs[0], qs[1], qs[2]))
        return _map_gate_name(gate_name), circuit

    return test_fixture


# New function for more complex gate structures:
def _generate_complex_gate_fixture(gate_sequence):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        qubits = [cirq.NamedQubit(f"q{i}") for i in range(len(gate_sequence))]
        for gate_op, qubit_indices in gate_sequence:
            gates_to_apply = [getattr(cirq, gate_op)(qubits[i]) for i in qubit_indices]
            circuit.append(gates_to_apply)
        return circuit

    return test_fixture


single_op_tests = [_fixture_name(s) for s in _one_qubit_gates]

# Generate three-qubit gate fixtures
for gate in _three_qubit_gates:
    name = _fixture_name(gate)
    locals()[name] = _generate_three_qubit_fixture(gate)


def _generate_measurement_fixture(gate_name: str):
    @pytest.fixture()
    def test_fixture():
        circuit = cirq.Circuit()
        q = cirq.NamedQubit("q")
        circuit.append(getattr(cirq, gate_name)(q))
        return _map_gate_name(gate_name), circuit

    return test_fixture


for gate in _measurements:
    name = _fixture_name(gate)
    locals()[name] = _generate_measurement_fixture(gate)

single_op_tests = [_fixture_name(s) for s in _one_qubit_gates]
rotation_tests = [_fixture_name(s) for s in _rotations]
double_op_tests = [_fixture_name(s) for s in _two_qubit_gates]
triple_op_tests = [_fixture_name(s) for s in _three_qubit_gates]
measurement_tests = [_fixture_name(s) for s in _measurements]



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
qBraid QIR setup file

"""
import os

from setuptools import setup


def read_requirements(file_path):
    """
    Reads a requirements file and returns the packages listed in it.

    Args:
        file_path (str): Path to the requirements file.

    Returns:
        list: List of package requirements.
    """
    with open(file_path, "r", encoding="utf-8") as file:
        return [line.strip() for line in file if line.strip()]


def read_version(file_path):
    """
    Extracts the version from a Python file containing a version variable.

    Args:
        file_path (str): Path to the Python file with the version variable.

    Returns:
        str: Version string, if found; otherwise, None.
    """
    with open(file_path, "r", encoding="utf-8") as file:
        return next(
            (
                line.split("=")[-1].strip().strip("\"'")
                for line in file
                if line.startswith("__version__")
            ),
            None,
        )


# Determine the directory where setup.py is located
here = os.path.abspath(os.path.dirname(__file__))

# Reading the package's version and requirements
version = read_version(os.path.join(here, "qbraid_qir/_version.py"))
install_requires = read_requirements(os.path.join(here, "requirements.txt"))
dev_requires = read_requirements(os.path.join(here, "requirements-dev.txt"))
docs_requires = read_requirements(os.path.join(here, "docs", "requirements.txt"))

setup(
    version=version,
    install_requires=install_requires,
    extras_require={"dev": dev_requires, "docs": docs_requires},
)
